#include "CSP.h"
#include "Assignment.h"
#include <cstdarg>
#include <climits>
#include <cstdio>

Assignment *Assignment::instance = 0;
map<vector<int>, int> sum;

/* Implementation of problem specific constraints */
bool notEqualCol(vector<int> variables) { 
     bool result;
     int init;
     vector<int> pair;

     for (int j = 0; unsigned(j) < variables.size(); j++) {
         pair.push_back(variables[j]);
         for (int k = 0; unsigned(k) < variables.size(); k++) {
             if (k < j) {
             init = -1;
             pair.push_back(variables[k]);
             for (int i = 0; unsigned(i) < pair.size(); i++) {
                 result = init != Assignment::a_instance()->getAssignment(pair[i]);
                 init = Assignment::a_instance()->getAssignment(pair[i]);
             }
             if (!result) return false;
             pair.pop_back();}
         }
         pair.pop_back();
     }

     return true;
}

bool notEqualRow(vector<int> variables) {
     int diffVar;
     int diffVal;
     vector<int> pair;

     for (int j = 0; unsigned(j) < variables.size(); j++) {
         pair.push_back(variables[j]);
         for (int k = 0; unsigned(k) < variables.size(); k++) {
             if (k < j) {
             diffVar = 0;
             diffVal = 0;
             pair.push_back(variables[k]);
             for (int i = 0; unsigned(i) < pair.size(); i++) {
                 diffVar = pair[i] - diffVar;
             }
             for (int i = 0; unsigned(i) < pair.size(); i++) {
                 diffVal = Assignment::a_instance()->getAssignment(pair[i]) - diffVal;
             }
             if (!((diffVar != diffVal) && (diffVar != -diffVal))) return false;
             pair.pop_back();}
         }
         pair.pop_back();
     }

     return true;
}

bool queenPlus(vector<int> variables) { 
     bool result;
     int init;
     vector<int> pair;

     for (int j = 0; unsigned(j) < variables.size(); j++) {
         pair.push_back(variables[j]);
         for (int k = 0; unsigned(k) < variables.size(); k++) {
             if (k < j) {
             init = -100;
             pair.push_back(variables[k]);
             for (int i = 0; unsigned(i) < pair.size(); i++) {
                 result = init != (pair[i] + Assignment::a_instance()->getAssignment(pair[i]));
                 init = pair[i] + Assignment::a_instance()->getAssignment(pair[i]);
             }
             if (!result) return false;
             pair.pop_back();}
         }
         pair.pop_back();
     }

     return true;
}

bool queenMinus(vector<int> variables) { 
     bool result;
     int init;
     vector<int> pair;

     for (int j = 0; unsigned(j) < variables.size(); j++) {
         pair.push_back(variables[j]);
         for (int k = 0; unsigned(k) < variables.size(); k++) {
             if (k < j) {
             init = -100;
             pair.push_back(variables[k]);
             for (int i = 0; unsigned(i) < pair.size(); i++) {
                 result = init != (Assignment::a_instance()->getAssignment(pair[i]) - pair[i]);
                 init = Assignment::a_instance()->getAssignment(pair[i]) - pair[i];
             }
             if (!result) return false;
             pair.pop_back();}
         }
         pair.pop_back();
     }

     return true;
}
/* ... */

vector<int> CSP::getDomainOf(int variable) {
	return varDomains[variable];
}

void CSP::addToDomain(int variable, vector<int> values) {	
	varDomains[variable] = values;
}

Assignment* CSP:: backTrackingSearch() {
        Assignment::a_instance()->reset();
        Assignment::a_instance()->setVarKeys(variables);
        /* Add here pre-given assignments if they exist */
	return recursiveBackTrackingSearch();
}

Assignment* CSP:: backTrackingSearch_MRV() {
        Assignment::a_instance()->reset();
        Assignment::a_instance()->setVarKeys(variables);
        /* Add here pre-given assignments if they exist */
	return recursiveBackTrackingSearch_MRV();
}

Assignment* CSP:: fowardCheckingSearch_MRV() {
        Assignment::a_instance()->reset();
        Assignment::a_instance()->setVarKeys(variables);
        /* Add here pre-given assignments if they exist */
	return recursiveFowardCheckingSearch_MRV();
}

Assignment* CSP::recursiveBackTrackingSearch() {
	int i;
        static int nodes = 0;
	
	if (Assignment::a_instance()->isComplete()) {
		cout << "Complete assignment" << endl;
                cout << nodes << " nodes expanded" << endl;
		return Assignment::a_instance();
	}

	int variable = Assignment::a_instance()->selectFirstUnassignedVariable();
	vector<int> domainValues = getDomainOf(variable);
	
	for (i = 0; i < int(domainValues.size()); i++) {
		int value = domainValues[i];

                nodes++;
                Assignment::a_instance()->setAssignment(variable, value);
		if (isSatisfiedWith(variable)) {
                        
			Assignment* result = recursiveBackTrackingSearch();
			if (result != NULL) {
				return result;
			}
			Assignment::a_instance()->remove(variable);
		}
                Assignment::a_instance()->remove(variable);
	}
	return NULL;	/* failure */
}

Assignment* CSP::recursiveBackTrackingSearch_MRV() {
        int i,j,count,mrv;
        static int nodes = 0;

        if (Assignment::a_instance()->isComplete()) {
                cout << "Complete assignment" << endl;
                cout << nodes << " nodes expanded" << endl;
                return Assignment::a_instance();
        }

        vector<int> vars = Assignment::a_instance()->selectAllUnassignedVariables();
        int mrv_variable = -1;

        mrv = getDomainOf(vars.front()).size()+1;
        vector<int> mrv_values;
        vector<int> temp;

        for (i = 0; unsigned(i) < vars.size(); i++) {
            count = 0;
            int variable = vars[i];
            vector<int> domainValues = getDomainOf(variable);
            temp.clear();
            for (j = 0; unsigned(j) < domainValues.size(); j++){
                Assignment::a_instance()->setAssignment(variable, domainValues[j]);
                if (isSatisfiedWith(variable)) {
                   count++;
                   temp.push_back(domainValues[j]);
                }
                  Assignment::a_instance()->remove(variable);
            }
            if (count < mrv) {
               mrv = count;
               mrv_variable = variable;
               mrv_values = temp;
            }
        }
        for (i = 0; unsigned(i) < vars.size(); i++) {
            count = 0;
            int variable = vars[i];
            vector<int> domainValues = getDomainOf(variable);
            temp.clear();
            for (j = 0; unsigned(j) < domainValues.size(); j++){
                Assignment::a_instance()->setAssignment(variable, domainValues[j]);
                if (isSatisfiedWith(variable)) {
                   count++;
                   temp.push_back(domainValues[j]);
                }
                  Assignment::a_instance()->remove(variable);
            }
            if (count < mrv) {
               mrv = count;
               mrv_variable = variable;
               mrv_values = temp;
            }
        }

        for (i = 0; unsigned(i) < mrv_values.size(); i++) {
                int value = mrv_values[i];

                Assignment::a_instance()->setAssignment(mrv_variable, value);
                        nodes++;
                        Assignment* result = recursiveBackTrackingSearch_MRV();
                        if (result != NULL) {
                                return result;
                        }
                        Assignment::a_instance()->remove(mrv_variable);
        }
        return NULL;    /* failure */
}


Assignment* CSP::recursiveFowardCheckingSearch_MRV() {
	int i,j,count,mrv;
        static map<int, vector<int> > Domains = varDomains;
        static int nodes = 0;
	
	if (Assignment::a_instance()->isComplete()) {
                varDomains = Domains;
                cout << "Complete assignment" << endl;
                cout << nodes << " nodes expanded" << endl;
                return Assignment::a_instance();
        }

	vector<int> Vars = Assignment::a_instance()->selectAllUnassignedVariables();
	int mrv_variable = -1;
	
        mrv = INT_MAX;
        for (i = 0; unsigned(i) < Vars.size(); i++) {
            count = 0;
            int variable = Vars[i];
        
            count = getDomainOf(variable).size();
            if (count < mrv) {
               mrv = count;
               mrv_variable = variable;
            }
        }
    
        vector<int> domainValues = getDomainOf(mrv_variable);
        map<int, vector<int> > oldVarDomains;	

	for (i = 0; unsigned(i) < domainValues.size(); i++) {
	    int value = domainValues[i];

            oldVarDomains = varDomains;
            Assignment::a_instance()->setAssignment(mrv_variable, value);
            vector<int> vars = getNeighborsOf(mrv_variable);
            if (vars.size() == 0) vars = Assignment::a_instance()->selectAllUnassignedVariables();
			
            for (j = 0; unsigned(j) < vars.size(); j++) {
                int variable = vars[j]; 
            
                if (!Assignment::a_instance()->hasAssignmentFor(variable)) {       
                   vector<int> var_domain = getDomainOf(variable);
                   while (!var_domain.empty() ) {
                         Assignment::a_instance()->setAssignment(variable, var_domain.back());
                         if (!isSatisfiedWith(variable)) {
                            vector<int> domainV = getDomainOf(variable);
                            vector<int> newValues;
                          
                            while (!domainV.empty() ) {
                                  if (domainV.back() != var_domain.back()) {
                                     newValues.push_back(domainV.back());
                                  }
                                  domainV.pop_back();
                            }
                             
                            varDomains[variable] = newValues;
                         }
                         Assignment::a_instance()->remove(variable);
                         var_domain.pop_back();
                   } 
                }
            } 
            nodes++; 
            Assignment* result = recursiveFowardCheckingSearch_MRV();
            if (result != NULL) {
               return result;
            }
            Assignment::a_instance()->remove(mrv_variable);
            varDomains = oldVarDomains;
       }
	
       return NULL;	
}


void CSP::addVariable(int var, int domSize) {
     
     vector<int> domain;
     
     for (int i = 1; i < domSize + 1; i++) {
		domain.push_back(i);
    }
     
     variables.push_back(var);
     addToDomain(var, domain);
}

void CSP::addConstraint(bool (*ptFunc)(vector<int>), int amount, ...) {
     vector<int> vars;
     vector< vector<int> > allVars;
     map<int, vector< vector<int> > > mappedVars;
     vector<bool(*)(vector<int>)> functs;
     bool flag;

     int val;
     int sval;
 
     va_list vl;
     va_start(vl,amount);
     for (int i=0;i<amount;i++)
     {
        val=va_arg(vl,int);
        sval = val;
        vars.push_back(sval);
        if (dictionary.find(sval) != dictionary.end()) {
           vector<bool(*)(vector<int>)> cons = dictionary[sval]; 
           flag = true;
           for (int j=0; unsigned(j) < cons.size(); j++) {
               if (cons[j]==ptFunc) {
                  flag = false;
                  break;
               }
           }
           if (flag) {
              dictionary[sval].push_back(ptFunc);
           } 
        }
        else {  
           functs.push_back(ptFunc);                      
           dictionary.insert ( pair<int, vector<bool(*)(vector<int>)> >(sval, functs) );
           functs.clear();
        }
     }
     va_end(vl);
 
     allVars.push_back(vars);
     
     
     if (constraints.find(ptFunc) != constraints.end()) {
        for (int i = 0; unsigned(i) < vars.size(); i++) {
              if (constraints[ptFunc].find(vars[i]) != constraints[ptFunc].end()) {
                 constraints[ptFunc][vars[i]].push_back(vars);
              }
              else {
                 constraints[ptFunc][vars[i]] = allVars;
              } 
         }   
     }
     else {  
         for (int i = 0; unsigned(i) < vars.size(); i++) {
              mappedVars[vars[i]] = allVars;
         }                      
         constraints.insert ( pair<bool(*)(vector<int>), map<int, vector<vector<int> > > >(ptFunc, mappedVars) );
     }
}

void CSP::addConstraint(bool (*ptFunc)(vector<int>), vector<int> vars) {
     vector< vector<int> > allVars;
     map<int, vector< vector<int> > > mappedVars;
     vector<bool(*)(vector<int>)> functs;
     bool flag;
 
     for (int i = 0; unsigned(i) < vars.size(); i++)
     {
        if (dictionary.find(vars[i]) != dictionary.end()) {
           vector<bool(*)(vector<int>)> cons = dictionary[vars[i]]; 
           flag = true;
           for (int j=0; unsigned(j) < cons.size(); j++) {
               if (cons[j]==ptFunc) {
                  flag = false;
                  break;
               }
           }
           if (flag) {
              dictionary[vars[i]].push_back(ptFunc);
           } 
        }
        else {  
           functs.push_back(ptFunc);                      
           dictionary.insert ( pair<int, vector<bool(*)(vector<int>)> >(vars[i], functs) );
           functs.clear();
        }
     }
 
     allVars.push_back(vars);
     
     
     if (constraints.find(ptFunc) != constraints.end()) {
        for (int i = 0; unsigned(i) < vars.size(); i++) {
              if (constraints[ptFunc].find(vars[i]) != constraints[ptFunc].end()) {
                 constraints[ptFunc][vars[i]].push_back(vars);
              }
              else {
                 constraints[ptFunc][vars[i]] = allVars;
              } 
         }   
     }
     else {  
         for (int i = 0; unsigned(i) < vars.size(); i++) {
              mappedVars[vars[i]] = allVars;
         }                      
         constraints.insert ( pair<bool(*)(vector<int>), map<int, vector<vector<int> > > >(ptFunc, mappedVars) );
     }
}

void CSP::addConstraint(bool (*ptFunc)(vector<int>), vector<int> vars,  int number) {
     vector< vector<int> > allVars;
     map<int, vector< vector<int> > > mappedVars;
     vector<bool(*)(vector<int>)> functs;
     bool flag;
 
     for (int i = 0; unsigned(i) < vars.size(); i++)
     {
        if (dictionary.find(vars[i]) != dictionary.end()) {
           vector<bool(*)(vector<int>)> cons = dictionary[vars[i]]; 
           flag = true;
           for (int j=0; unsigned(j) < cons.size(); j++) {
               if (cons[j]==ptFunc) {
                  flag = false;
                  break;
               }
           }
           if (flag) {
              dictionary[vars[i]].push_back(ptFunc);
           } 
        }
        else {  
           functs.push_back(ptFunc);                      
           dictionary.insert ( pair<int, vector<bool(*)(vector<int>)> >(vars[i], functs) );
           functs.clear();
        }
     }
 
     allVars.push_back(vars);
     
     
     if (constraints.find(ptFunc) != constraints.end()) {
        for (int i = 0; unsigned(i) < vars.size(); i++) {
              if (constraints[ptFunc].find(vars[i]) != constraints[ptFunc].end()) {
                 constraints[ptFunc][vars[i]].push_back(vars);
              }
              else {
                 constraints[ptFunc][vars[i]] = allVars;
              } 
         }   
     }
     else {  
         for (int i = 0; unsigned(i) < vars.size(); i++) {
              mappedVars[vars[i]] = allVars;
         }                      
         constraints.insert ( pair<bool(*)(vector<int>), map<int, vector<vector<int> > > >(ptFunc, mappedVars) );
     }
     
     sum[vars] = number;
}

bool CSP::isSatisfiedWith(int variable) {
     vector< vector<int> >::iterator itvs;
     map<int, vector<bool(*)(vector<int>)> >::iterator dict;

     bool flag = true;

     if (dictionary[variable].size() == 0) {return true;}
     //if (dictionary.find(variable) == dictionary.end()) {return true;} 

     vector<bool(*)(vector<int>)> cons = dictionary[variable];
     for (int i=0; unsigned(i) < cons.size(); i++) {
         for (itvs = constraints[cons[i]][variable].begin(); itvs < constraints[cons[i]][variable].end(); itvs++ ) {
             flag = true;
             for ( int j = 0 ; unsigned(j) < (*itvs).size() ; j++) {
                 if ((*itvs)[j] == variable) continue; 
                 if (!Assignment::a_instance()->hasAssignmentFor((*itvs)[j])) {
                    flag = false;
                    break;
                 }	     
             }
             if (flag) {
                if(cons[i](*itvs)!=true) return false ;
             }  
         }
     }
     return true;
} 

void CSP::makeNeighbors(int variable, int nVariable) {
     neighbors[variable].push_back(nVariable);
     neighbors[nVariable].push_back(variable);
}

vector<int> CSP::getNeighborsOf(int variable) {
     return neighbors[variable];
}

void CSP::f() {
     map<bool(*)(vector<int>), map<int, vector<vector<int> > > >::iterator iterc;
     map<int, vector<vector<int> > >::iterator iter;
     map<int, vector<int> >::iterator it;

     cout << "---------------------" << endl;
     cout << "Neighbors" << endl;
     for (it = neighbors.begin(); it != neighbors.end(); it++) {
         cout << "\nVariable " << it->first << " has neighbors: ";
         for (int i = 0; unsigned(i) < it->second.size(); i++ ) {
             cout << it->second[i] << " ";
         }
     }
     cout << endl; 

     cout << "---------------------" << endl;
     cout << "Dictionary" << endl;
     for (int i = 0; unsigned(i) < variables.size(); i++) {
         cout << "\nVariable " << variables[i] << " is in " << dictionary[variables[i]].size() << " constraints";
     }
     cout << endl;

     cout << "---------------------" << endl;
     cout << "Constraints" << endl;
     int i;

     for (i = 0, iterc = constraints.begin(); iterc != constraints.end(); i++, iterc++) {
         cout << "\nConstraint No. " << i+1 << " has variables";
         for (iter = constraints[iterc->first].begin(); iter != constraints[iterc->first].end(); iter++) {
             cout << "\n" << iter->first << " in constrained variable sets: ";
             cout << "{";
             for (int j=0; unsigned(j) < iter->second.size(); j++) {
                 cout << "{ ";
                 for (int k=0; unsigned(k) < iter->second[j].size(); k++) {
                     cout << iter->second[j][k] << " ";
                 }
                 cout << "}";
             }
             cout << "}"; 
         }
         cout << endl;
     }
}
     
     
