#ifndef _CSP_H_
#define _CSP_H_
#include "Assignment.h"
#include <vector>
#include <map>
#include <iostream>
#include <cstdarg>

using namespace std;

/* Define your own constraints that fit the problem */
bool notEqualRow(vector<int>);
bool notEqualCol(vector<int>);
bool queenPlus(vector<int>);
bool queenMinus(vector<int>);
/* ... */

class CSP {
    private:
        Assignment* recursiveBackTrackingSearch();
        Assignment* recursiveBackTrackingSearch_MRV(); 
        Assignment* recursiveFowardCheckingSearch_MRV();   
        map<int, vector<int> > varDomains;
        vector<int> variables; 
        map<int, vector<int> > neighbors;   
        map<bool(*)(vector<int>), map<int, vector<vector<int> > > > constraints;
        map<int, vector<bool(*)(vector<int>)> > dictionary;
	public:
		vector<int> getDomainOf(int variable);
		void addToDomain(int variable, vector<int> values);
		Assignment* backTrackingSearch();
                Assignment* backTrackingSearch_MRV();
                Assignment* fowardCheckingSearch_MRV(); 
		void addVariable(int var, int domSize);
		void addConstraint(bool (*ptFunc)(vector<int>), int, ...);
                void addConstraint(bool (*ptFunc)(vector<int>), vector<int>);
                void addConstraint(bool (*ptFunc)(vector<int>), vector<int>, int);
		bool isSatisfiedWith(int variable); 
                void makeNeighbors(int variable, int nVariable);
                vector<int> getNeighborsOf(int variable);
                void f();    
};

#endif
