#include <iostream>
#include <string>
#include "CSP.h"
#include "Assignment.h"
#include <cstdarg>
#include <sstream>
#include <cstdlib>

using namespace std;

int main(int argc, char* argv[])
{    
    try  {
		
	int numberOfQueens = (argc > 1) ? atoi(argv[1]) : 8;
        string choice;
        int ichoice;
        bool flag;
	
	/* Create the instance of a CSP problem */
        CSP csp;
	
        /* Method 1 */

	/* Add the variables with their domains for the N-Queens problem */ 
        
        for (int i = 1; i < numberOfQueens + 1; i++) {
		csp.addVariable(i, numberOfQueens); 
	}
        
	/* Add the constraints for the problem */
        
        for (int i = 1; i < numberOfQueens + 1; i++) 
            for (int j = 1; j < numberOfQueens + 1; j++) {
                if (j < i) {
                   csp.addConstraint(notEqualCol, 2, i, j);
                   csp.addConstraint(notEqualRow, 2, i, j);
                }
            }
        
        /* Method 2 */
        
        /* Add the variables with their domains for the N-Queens problem */
        /*
        for (int i = 1; i < numberOfQueens + 1; i++) {
		csp.addVariable(i, numberOfQueens); 
	} 
        */

        /* Add the constraints for the problem */
        /*
        for (int i = 1; i < numberOfQueens + 1; i++) 
            for (int j = 1; j < numberOfQueens + 1; j++) {
                if (j < i) {
                   csp.addConstraint(notEqualCol, 2, i, j);
                   csp.addConstraint(queenPlus, 2, i, j);
                   csp.addConstraint(queenMinus, 2, i, j);
                }
            }
        */
 
        /* Print information about neighbours, dictionary and constraints */
        csp.f();

	cout << "---------------------" << endl;
	cout << "CSP problem - Backtracking " << endl;

        Assignment *a; 

        /* Solve the problem */
        do {
           flag = true;
           cout << "Choose solving method:\n" << 
                   "1. Chronological BackTracking\n" <<
                   "2. Chronological BackTracking + MRV heuristic\n" <<
                   "3. ForwardChecking + MRV heuristic\n" <<
                   "4. Exit" << endl;
          
           choice = "";
           cin >> choice;
           istringstream buffer(choice);
           buffer >> ichoice;
          
           switch (ichoice) {
                 case 1: a = csp.backTrackingSearch(); flag = false; break;
                 case 2: a = csp.backTrackingSearch_MRV(); flag = false; break;
                 case 3: a = csp.fowardCheckingSearch_MRV(); flag = false; break;
                 case 4: return 0;
                 default: cout << "Invalid option, please try again" << endl; continue;
           }
        } while (flag);
	
        /* Print the solution if it exists */
	if (a == NULL) {
	   cout << "CSP has no solution!" << endl;
           return 0;
	} 
        else {
           cout << endl;
	   a->print();
	}
	
	/* pretty N-Queens print */
	cout << "---------------------" << endl;
        cout << "N-queens problem" << endl;
        cout << "\n " ; 
        for (int i = 1; i < (numberOfQueens*4); i++) 
            cout << "-" ;
        cout << endl;
        for (int i = 1; i < numberOfQueens + 1; i++) {
             cout << "|" ;
             for (int j = 1; j < numberOfQueens + 1; j++) {
                 int row = a->getAssignment(i);
                 (row == j) ? cout << " Q |" : cout << "   |" ;
             }
             cout << endl;
             if (i != numberOfQueens) {
                cout << " " ;
                for (int k = 1; k < (numberOfQueens*4); k++)
                    cout << "-" ;
                cout << endl;
             }
        }
        cout << " " ;
        for (int i = 1; i < (numberOfQueens*4); i++)
            cout << "-" ;
        cout << endl;
              
	cout << "---------------------" << endl;

	return 0;
	
	}  catch (exception& exc)  {

	cerr << exc.what() << "\n";

    }  catch (...)  {

	cerr << "Unknown exception" << "\n";
    }
}
